from typing import Tuple, List
from PIL import Image
import fast_deskew
import cv2
import numpy as np


def deskew(image: str) -> Tuple[Image.Image, float]:
    _, best_angle = fast_deskew.deskew_image(image, False)
    img = Image.open(image)
    return img.rotate(best_angle), best_angle


def rotatebox(bbox: List[List[int]], image: Image.Image, angle: float):
    # https://gist.githubusercontent.com/Joanne03/5941a9b4db4fa7c652a2d7f67b11a09b/raw/f99b2f18ad0fbe680ac631e531782b957f158def/rotate_bbox.py
    height, width = image.size
    image_center_x, image_center_y = width // 2, height // 2

    rotated_bbox = []

    for i, coord in enumerate(bbox):
      rot_matrix = cv2.getRotationMatrix2D((image_center_x, image_center_y), angle, 1.0)
      cosinus, sinus = abs(rot_matrix[0, 0]), abs(rot_matrix[0, 1])
      new_width = int((height * sinus) + (width * cosinus))
      new_height = int((height * cosinus) + (width * sinus))
      rot_matrix[0, 2] += (new_width / 2) - image_center_x
      rot_matrix[1, 2] += (new_height / 2) - image_center_y
      v = [coord[0], coord[1], 1]  # ?
      adjusted_coord = np.dot(rot_matrix, v)
      rotated_bbox.append((int(adjusted_coord[0]), int(adjusted_coord[1])))

    return rotated_bbox



from typing import List, Dict
from ultralytics import YOLO
from ultralytics.engine.results import Results

from yaltai.preprocessing import deskew, rotatebox


def segment(
        model: str,
        device: str,
        input: str,
        apply_deskew: bool = False,
        max_angle: float = 10.0
) -> Dict[str, List[List[int]]]:
    """

    Returns {
        cls_name: [
            [[x0, y0], [x1, y0], [x1, y1], [x0, y1], [x0, y0]]
        ]
    }
    """
    model = YOLO(model)
    model.to(device)
    rotated_input = None
    angle = 0
    predictions: List[Results] = []
    if apply_deskew:
        rotated_input, angle = deskew(input)
        if abs(angle) > max_angle:
            predictions = model.predict(input, save=False)
            rotated_input = None
        else:
            predictions = model.predict(rotated_input, save=False)
    else:
        predictions = model.predict(input, save=False)

    names: List[str] = list(set([
        name
        for res in predictions
        for name in res.names.values()
    ]))

    out = {
        name: []
        for name in names
    }
    for pred in predictions:
        for box, cls_id in zip(pred.boxes.xyxy, pred.boxes.cls):
            cls_name = pred.names[cls_id.item()]
            x0, y0, x1, y1 = box.tolist()
            points = [[x0, y0], [x1, y0], [x1, y1], [x0, y1], [x0, y0]]

            if apply_deskew and rotated_input is not None:
                points = rotatebox(points, rotated_input, -angle)
                points.append(points[0])
            out[cls_name].append(points)

    return out



from typing import Optional, Callable, Union, List, Dict, Any

import PIL
import logging
import numpy as np
from kraken.blla import compute_segmentation_map, vec_regions, vec_lines

from kraken.lib.segmentation import (polygonal_reading_order, scale_regions)
from kraken.lib import vgsl
from kraken.lib.exceptions import KrakenInvalidModelException
from kraken.lib.util import get_im_str

logger = logging.getLogger(__name__)


def segment(im: PIL.Image.Image,
            text_direction: str = 'horizontal-lr',
            mask: Optional[np.ndarray] = None,
            reading_order_fn: Callable = polygonal_reading_order,
            model: Union[List[vgsl.TorchVGSLModel], vgsl.TorchVGSLModel] = None,
            device: str = 'cpu',
            regions: Optional[Dict[str, List[List[int]]]] = None,
            ignore_lignes: bool = False) -> Dict[str, Any]:
    r"""
    Segments a page into text lines using the baseline segmenter.

    Segments a page into text lines and returns the polyline formed by each
    baseline and their estimated environment.

    Args:
        im: Input image. The mode can generally be anything but it is possible
            to supply a binarized-input-only model which requires accordingly
            treated images.
        text_direction: Passed-through value for serialization.serialize.
        mask: A bi-level mask image of the same size as `im` where 0-valued
              regions are ignored for segmentation purposes. Disables column
              detection.
        reading_order_fn: Function to determine the reading order.  Has to
                          accept a list of tuples (baselines, polygon) and a
                          text direction (`lr` or `rl`).
        model: One or more TorchVGSLModel containing a segmentation model. If
               none is given a default model will be loaded.
        device: The target device to run the neural network on.
        regions: Regions computed using YOLOv8

    Returns:
        A dictionary containing the text direction and under the key 'lines' a
        list of reading order sorted baselines (polylines) and their respective
        polygonal boundaries. The last and first point of each boundary polygon
        are connected.

        .. code-block::
           :force:

            {'text_direction': '$dir',
             'type': 'baseline',
             'lines': [
                {'baseline': [[x0, y0], [x1, y1], ..., [x_n, y_n]], 'boundary': [[x0, y0, x1, y1], ... [x_m, y_m]]},
                {'baseline': [[x0, ...]], 'boundary': [[x0, ...]]}
              ]
              'regions': [
                {'region': [[x0, y0], [x1, y1], ..., [x_n, y_n]], 'type': 'image'},
                {'region': [[x0, ...]], 'type': 'text'}
              ]
            }

    Raises:
        KrakenInvalidModelException: if the given model is not a valid
                                     segmentation model.
        KrakenInputException: if the mask is not bitonal or does not match the
                              image size.
    """
    if isinstance(model, vgsl.TorchVGSLModel):
        model = [model]

    for nn in model:
        if nn.model_type != 'segmentation':
            raise KrakenInvalidModelException(f'Invalid model type {nn.model_type} for {nn}')
        if 'class_mapping' not in nn.user_metadata:
            raise KrakenInvalidModelException(f'Segmentation model {nn} does not contain valid class mapping')

    im_str = get_im_str(im)
    logger.info(f'Segmenting {im_str}')

    if ignore_lignes:
        return {'text_direction': text_direction,
                'type': 'baselines',
                'lines': [],
                'regions': regions,
                'script_detection': False}

    for net in model:
        if 'topline' in net.user_metadata:
            loc = {None: 'center',
                   True: 'top',
                   False: 'bottom'}[net.user_metadata['topline']]
            logger.debug(f'Baseline location: {loc}')

        rets = compute_segmentation_map(im, mask, net, device)

        # We can't clear the heatmap of regions because it would mess up

        if "regions" in rets:
            del rets["regions"]

        # flatten regions for line ordering/fetch bounding regions
        line_regs = []
        suppl_obj = []
        for cls, regs in regions.items():
            line_regs.extend(regs)
            if rets['bounding_regions'] is not None and cls in rets['bounding_regions']:
                suppl_obj.extend(regs)

        # convert back to net scale
        suppl_obj = scale_regions(suppl_obj, 1/rets['scale'])
        line_regs = scale_regions(line_regs, 1/rets['scale'])
        lines = vec_lines(**rets,
                          regions=line_regs,
                          reading_order_fn=reading_order_fn,
                          text_direction=text_direction,
                          suppl_obj=suppl_obj,
                          topline=net.user_metadata['topline'] if 'topline' in net.user_metadata else False)

    if len(rets['cls_map']['baselines']) > 1:
        script_detection = True
    else:
        script_detection = False

    return {'text_direction': text_direction,
            'type': 'baselines',
            'lines': lines,
            'regions': regions,
            'script_detection': script_detection}
